﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Net;
using System.Net.Sockets;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Security.Cryptography;
using Lab2DiffieHellman;
using System.IO;

namespace Lab2DHClient
{
    public class NetClient
    {
        /// <summary>
        /// This method starts client application, which connects to server, exchanges the key with the server and starts listening for messages.
        /// </summary>
        /// <param name="serverIP">Valid IPv4 address of the server.</param>
        /// <param name="serverPort">Port number where the server is listening.</param>
        public static void StartClient(IPAddress serverIP, int serverPort, bool usePrecounted, int keyLength)
        {
            try
            {
                IPEndPoint remoteEP = new IPEndPoint(serverIP, serverPort);

                TcpClient client = new TcpClient();
                DifHel.WriteLine("Trying to establish the connection.");
                client.Connect(serverIP, serverPort);
                NetworkStream ns = client.GetStream();
                DifHel.WriteLine("Connection established.");

                byte[] key = ExchangeKey(ns, usePrecounted, keyLength);

                while (true)
                {
                    DifHel.WriteLine("Type in the message: ('end' for turning off)");
                    string message = Console.ReadLine();
                    if (message == "end") break;
                    SendMessage(ns, key, message);

                    string received = GetReply(ns, key);

                    DifHel.WriteLine(received);
                }
                DifHel.WriteLine("Going to turn off now.");
            }
            catch (System.IO.IOException)
            {
                DifHel.WriteLine("The connection to the server was lost. Lets try it once again.");
            }
            catch (System.Net.Sockets.SocketException)
            {
                DifHel.WriteLine("The connection to the server was not established. Lets try it once again.");
            }
            catch (Exception e)
            {
                DifHel.WriteLine(e.ToString());
            }
        }

        /// <summary>
        /// THis method is used for encrypting and sending the message to the server.
        /// </summary>
        /// <param name="ns">NetworkStream to the server.</param>
        /// <param name="key">Exchanged private key.</param>
        /// <param name="message">Message to be delivered to the other side.</param>
        private static void SendMessage(NetworkStream ns, byte[] key, string message)
        {
            byte[] outgoing;

            using (Aes encryption = Aes.Create())
            {
                encryption.Key = key;
                encryption.GenerateIV();

                byte[] mess = AesTools.EncryptStringToBytes(message, encryption);
                outgoing = new byte[encryption.IV.Length + mess.Length];
                encryption.IV.CopyTo(outgoing, 0);
                mess.CopyTo(outgoing, encryption.IV.Length);
            }

            try
            {
                ns.Write(outgoing, 0, outgoing.Length);
                DifHel.WriteLine(string.Format("Succesfully sent: {0}", message));
            }
            catch (Exception e)
            {
                DifHel.WriteLine(e.ToString());
            }
        }

        /// <summary>
        /// This method is used for receiving response from server, deciphering it and presenting it to the user.
        /// </summary>
        /// <param name="ns">NetworkStream to the server.</param>
        /// <param name="key">Exchanged private key.</param>
        /// <returns>String with received reply.</returns>
        private static string GetReply(NetworkStream ns, byte[] key)
        {
            byte[] incoming = new byte[1024];
            int bytesRead = ns.Read(incoming, 0, incoming.Length);
            byte[] trimmed = new byte[bytesRead - (16)];
            byte[] iv = new byte[16];

            for (int i = 0; i < iv.Length; i++)
            {
                iv[i] = incoming[i];
            }

            for (int i = iv.Length; i < bytesRead; i++)
            {
                trimmed[i - iv.Length] = incoming[i];
            }
            string received = "";
            using (Aes encryption = Aes.Create())
            {
                encryption.Key = key;
                encryption.IV = iv;

                received = AesTools.DecryptStringFromBytes(trimmed, encryption);
            }
            return received;
        }

        /// <summary>
        /// Method used for exchanging the key after establishing connection with the server.
        /// </summary>
        /// <param name="ns">NetworkStream to the server.</param>
        /// <param name="keyLength">Length of key in bits. Please note that if the usePrecaunted parameter is set to true, the keyLenght is always default (2048).</param>
        /// <param name="usePrecounted">If true, the system uses predefined p and g values instead of freshly calculated ones. This keys has harcoded length of 2048 bits.</param>
        /// <returns>Exchanged secret key.</returns>
        private static byte[] ExchangeKey(NetworkStream ns, bool usePrecounted, int keyLength)
        {
            DifHel DHClient = new DifHel(keyLength, usePrecounted);

            try
            {
                ns.Write(Encoding.Unicode.GetBytes(DHClient.ToString()), 0, Encoding.Unicode.GetByteCount(DHClient.ToString()));
            }
            catch (Exception e)
            {
                DifHel.WriteLine(e.ToString());
            }

            byte[] incoming = new byte[8192];
            int bytesRead = ns.Read(incoming, 0, incoming.Length);

            DHClient.ParseReply(Encoding.Unicode.GetString(incoming));
            DifHel.WriteLine(string.Format("The key was succesfully exchanged. Its length is {0} bytes.", DHClient.Key.Length));
            return DHClient.Key;
        }


    }
}
